/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>
#include "vector.h"
#include "vty.h"
#include "command.h"
#include "getopt.h"
#include "thread.h"

#include "stream.h"
#include "network.h"
#include "prefix.h"
#include "version.h"
#include "memory.h"
#include "assert.h"

#include "if.h"
#include "sockunion.h"

#include "log.h"
#include "pim.h"
#include "pimd.h"
#include "pim_fsm.h"
#include <netinet/igmp.h>

#include <netinet/ip_mroute.h>
#include "tables.h"
#include "pim_route.h"
#include "pim_packet.h"
#include "igmp_def.h"
#include "if_idx_nexthop.h"
#include "pim_joinprune.h"

#include "zbug.h"
ZBUG_EXTERN

extern int igmp_sock;
extern int in_cksum (void *ptr, int nbytes);
extern  list mgrp_hash[];

extern char b1[16], b2[16], b3[16], b4[16];  /* For inet_2a */
extern struct best_route best_route;

char igmp_sendbuf[4096];

int
query_thread(struct thread *t_query)
{
  struct pim_interface *pi = THREAD_ARG (t_query);
  struct interface *ifp;
  u_int32_t timer;
  ifp = pi->ifp;

  if( pi->Startup_Query_Count > 0 ) {
    timer =  pi->Startup_Query_Interval;
    pi->Startup_Query_Count--;
  }
  else timer =  pi->Query_Interval;

  igmp_query_neighbors(ifp, 0, pi->Query_Response_Interval, 
		       INADDR_ALLHOSTS_GROUP);
  t_query = thread_add_timer(master, query_thread, pi, timer);
  pi->t_query = t_query;
  return 0;
}

int 
other_querier_present(struct thread *t_other_querier_present)
{
  /* The querier is gone. We have to do the work */

  struct pim_interface *pi = THREAD_ARG(t_other_querier_present);
  struct interface *ifp;
  ifp = pi->ifp;
  pi->t_other_querier_present = NULL;
  pi->flags |= IF_IGMP_QUERIER;
  pi->t_query = thread_add_timer(master, query_thread, pi, 0);
  return 0;
}

int
stop_query(struct pim_interface *pi)
{
  struct interface *ifp;
  u_int32_t timer;
  
  zassert(pi, -1);
  ifp = pi->ifp;

  pi->flags &= ~IF_IGMP_QUERIER;
  if( pi->t_query ) {
    thread_cancel(pi->t_query);
  pi->t_query = NULL;
  }
  timer = pi->Other_Querier_Present_Interval;
  pi->t_other_querier_present = thread_add_timer(master,
			other_querier_present, pi, timer);

  pi->Startup_Query_Count = 2; /* pi->Startup_Query_Count_conf;*/
  return 0;
}

int
igmp_if_start(struct interface *ifp)
{
  struct pim_interface *pi;
  zassert(ifp, -1);
  pi = ifp->info;
  
  init_igmp_timers(pi);
  pi->igmp_static_group_list= list_init ();

  pi->t_query = thread_add_timer(master, query_thread, pi, 0);
  return 0;
}

int
igmp_if_stop(struct interface *ifp)
{
  struct pim_interface *pi;
  listnode node;
  u_int32_t g;

  zassert(ifp, -1);
  pi = ifp->info;

  if(pi->t_query) {
    thread_cancel(pi->t_query);
    pi->t_query =NULL;
  }

  /* Remove static IGMP groups joined */

  
  for (node=listhead (pi->igmp_static_group_list); node; nextnode (node)) {
    g = (u_int32_t) getdata (node);
  }
  list_delete_all (pi->igmp_static_group_list);
  pi->igmp_static_group_list = NULL;
  return 0;
}

/* Check if interface is local */
int
local_in_addr(struct in_addr *src)
{
  struct interface *ifp;
  struct pim_interface *pi;
  listnode node;

  for (node=listhead (iflist); node; nextnode (node)) {
    ifp = getdata (node);
    pi=ifp->info;
    if( ! pi->address ) continue;
    if( pi->address->u.prefix4.s_addr == src->s_addr ) return 1;
  }
  return 0;
}

int
Version_1_Router_Present(struct thread *t_Version_1_Router_Present)
{
  struct pim_interface *pi = THREAD_ARG (t_Version_1_Router_Present);
  //struct interface *ifp;
  
  /* Our IGMPV1 querier has timed out. We turn into V2 */

  pi->t_Version_1_Router_Present = NULL;
  pi->flags &= ~IF_IGMP_V1;
  return 0;
}

void 
parse_igmp_query(struct in_addr *src, struct igmphdr *igmp, 
		 struct interface *ifp)
{
  struct pim_interface *pi = ifp->info;
  //listnode node;
 
  if(local_in_addr(src)) {
    printf("Query from own interface %s\n", ifp->name);
    return;
  }

  /* IGMPV2 querier selection */

  if(igmp->code != 0 && pi->flags & IF_IGMP_QUERIER &&  
     src->s_addr < pi->address->u.prefix4.s_addr) {

    /* We not allowed to do Querier into non-Querier when
       a leave process is active. In our implementation 
       this should not be a problem. We test for it anyway */

    if(pi->igmp_leave_process == FALSE) stop_query(pi);
    return;
  }

  if(! (pi->flags & IF_IGMP_V1) ) {
    if(igmp->code == 0) {

      /* An IGMPV1 QUERY when in V2 mode */
      /* We must warn according to RFC 2236 */

      /* TODO: Put ratelimit here */
      printf("RX IGMP V1 QUERY on interface %s from %s while in V2 mode. ",
	     ifp->name, inet_ntoa(*src));
      printf("Please configure routers properly\n");
    }
  }
  
  if(igmp->code == 0) {
    /* Turn to mode IGMPV1 */
    pi->flags |= IF_IGMP_V1;

    /* Start a supervision timer */
    pi->t_Version_1_Router_Present =
      thread_add_timer(master, Version_1_Router_Present, pi, 
		       pi->Version_1_Router_Present_Timeout);

    /* Now for us to become V1 querier we have to use the DR information
       from PIM */

    if(pi->flags & IF_DR) {
      /* Were we in V2 before */
      if(! (pi->flags & IF_IGMP_V1) ) {
      }

      /* TODO: Fix this mess */

    }
    pi->flags |= IF_IGMP_QUERIER;
  }
  else pi->flags &= ~IF_IGMP_V1; 
}

int
oif_list_in_use(list oif_list, int flag)
{
  listnode node;
  struct oif *mi;
  
  for (node = listhead (oif_list); node; nextnode (node)) {
    mi = getdata(node);
    if(mi->flags & flag)
      return 1;
  }
  return 0;
}

void
igmp_oif_remove(  struct oif *mi )
{
  struct interface *ifp;
  struct mgrp *mg;
  
  mg = mi->mg;
  zassert(mg, -1);
  
  ifp = mi->ifp;
  zassert(ifp, -1);
  
//  mgrp_oif_modify(mg, ifp->ifindex, "Del");

  /* Remove multicast group from group-list */

  if(mi->flags & MRT_PIM_IN_USE)
    mi->flags &= ~MRT_IGMP_IN_USE;
  else
    mg_oif_list_delete (mi);
  
  /* Any igmp interfaces still left? */
  if(!oif_list_in_use(mg->oif_list, MRT_IGMP_IN_USE))
    {
      /* if not clear IGMP_IN_USE */
      mg->flags &= ~MRT_IGMP_IN_USE;
    }  
  
  if( list_isempty(mg->oif_list) ) {
    if(mg->flags & MRT_PIM_IN_USE)
      {
	mg->flags |= MRT_PRUNED;
	mg->flags &= ~MRT_IGMP_IN_USE;
      }
    else
      mg_remove(mg);
  }
}

int
igmp_thread_timeout(struct thread *t_igmp_timeout)
{
  struct oif *mi = THREAD_ARG( t_igmp_timeout );
  mi->igmp.t_igmp_timeout=NULL;
  ZBUG_G(mi->mg->group)
    {
      fprintf(ZBUG_LOG, "IGMP oif timeout %s ", mi->ifp->name);
      pim_flags_f(ZBUG_LOG, 0, mi->mg->group, mi->flags);
      fflush(ZBUG_LOG);
    }
  igmp_oif_remove(mi);
  return 1;
}

int 
send_group_specific_query(struct thread *t_send_group_specific_query)
{
  u_int32_t timer;
  struct oif *mi = THREAD_ARG(t_send_group_specific_query);
  struct interface *ifp;
  struct pim_interface *pi;
  struct mgrp *mg;

  zassert(mi, -1);
  ifp = mi->ifp;
  zassert(ifp, -1);
  pi=ifp->info;
  zassert(mi, -1);
  mg = mi->mg;
    
  if( mi->igmp.leave_count == pi->Last_Member_Query_Count) {

    mi->igmp.leave_count = 0;
    pi->igmp_leave_process = FALSE; /* Set IGMP leave process idle */

    /* Have we received any report? */
    if(mi->igmp.exptime_igmp != mi->igmp.leave_request_time) return 0;

    /* Otherwise
      
       1) Delete interface from group 
       2) And maybe delete the group if needed */

    if(mi->igmp.t_igmp_timeout) {
      thread_cancel(mi->igmp.t_igmp_timeout);
      mi->igmp.t_igmp_timeout = NULL;
    }
    
    printf("Removing interface %s from group %s\n",
	   ifp->name,
	   inet_2a(mg->group, b1));

    igmp_oif_remove( mi );
    
    return 0;
  }
  igmp_query_neighbors(ifp, mg->group, 
		       pi->Last_Member_Query_Interval,
		       mg->group);
  mi->igmp.leave_count++;
  timer = pi->Last_Member_Query_Interval;
  pi->t_query = thread_add_timer(master, send_group_specific_query, mi, 
				 timer);
  return 0;
}

int
parse_igmp_leave(struct in_addr *src, struct igmphdr *igmp, 
		 struct interface *ifp)
{
  struct pim_interface *pi;
  struct mgrp *mgrp;
  struct oif *mi;
  listnode node;

  zassert(ifp, -1);
  pi = ifp->info;

  if( !(pi->flags & IF_IGMP_QUERIER)) return 0; /* Not for us */

  if(local_in_addr(src)) {
    printf("Report from own interface %s\n", ifp->name);
    return 0;
  }
  printf("IGMP leave from %s %s\n", inet_2a(igmp->group, b1), 
	 inet_2a(src->s_addr, b2) );

  /* Lookup group first */

  if((mgrp=mg_lookup(igmp->group)) == NULL ) {
    printf("Error no group %s\n", inet_2a(igmp->group, b1));
    return 0; 
  }

  /* Lookup joined interface */

  for (node = listhead (mgrp->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp) {

      /* Send group specific query for group with 
	 timer Last_Member_Query_Interval */

      if(mi->igmp.leave_count) return 0; /* We are already processing leave */
      else mi->igmp.leave_count = TRUE;

      pi->igmp_leave_process = TRUE; /* Set IGMP leave process active */

      mi->igmp.leave_request_time = mi->igmp.exptime_igmp;
      igmp_query_neighbors(ifp, igmp->group, 
			   pi->Last_Member_Query_Interval,
			   igmp->group);

      /* And Last_Member_Query_Interval should be used in the timer 
	 as well */

      pi->t_query = thread_add_timer(master, send_group_specific_query, 
				     mi,  pi->Last_Member_Query_Interval);

      return 0;
    }
  } 
  printf("Error no joined interface for group %s\n", 
	 inet_2a(igmp->group, b1));
  return 0;
}

int
process_igmp_report(struct in_addr *src, u_int32_t group,
		 struct interface *ifp, u_int32_t permanent)
{
  struct pim_interface *pi;
  struct mgrp *mg;
  struct oif *mi;
  listnode node;
  int ret;
  zassert(ifp->info, -1);
  pi = ifp->info;

  /* Ignore LAN groups not for routing  */

  if(ntohl(group) <= INADDR_MAX_LOCAL_GROUP) return 0;
  /*  if(LOCAL_MCAST(ntohl(group)) return 0; */

  if((mg=mg_lookup(group)) == NULL ) {
    mg = pim_mg_add(group, 0); /* Startup valuse. Check this */
  }

  ZBUG(ZBUG_IGMP_MESSAGES) 
    printf("***Report for interface %s for group %s\n", 
	    ifp->name, inet_2a(group, b1));

  zassert(mg, -1);
  for (node = listhead (mg->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp) {

      if( mi->igmp.t_igmp_timeout != NULL ) {
	thread_cancel(mi->igmp.t_igmp_timeout);
	mi->igmp.t_igmp_timeout = NULL;
      }
      if(! permanent) {
	mi->igmp.exptime_igmp = time(NULL) + pi->Group_Membership_Interval;
	mi->igmp.t_igmp_timeout = thread_add_timer(master, igmp_thread_timeout,
						   mi, pi->Group_Membership_Interval);
      }
//      mg->exptime = time(NULL) + pi->Group_Membership_Interval;
      mi->igmp.last_reporter = src->s_addr;
      if( mi->uptime == 0)  mi->uptime = time(NULL);
      mg->flags |= MRT_IGMP_IN_USE;
      mi->flags |= MRT_IGMP_IN_USE;
      return 0;
    }
  }
  

  /* Create an oif */

  mi = oif_alloc (MTYPE_PIM_OIF);
  list_add_node(mg->oif_list, mi);
  mi->ifp = ifp;
  mi->igmp.last_reporter = src->s_addr;
  mi->uptime = time(NULL);
  mi->igmp.exptime_igmp = time(NULL) + pi->Group_Membership_Interval;

  if(permanent) {
    mi->igmp.exptime_igmp = 0;
    ret=if_group (ifp, igmp_sock, htonl(group), IP_ADD_MEMBERSHIP);
    if(ret < 0) printf("Error joining permanent group\n");
  }
  else {
      mi->igmp.t_igmp_timeout = thread_add_timer(master, igmp_thread_timeout,
						 mi, pi->Group_Membership_Interval);
  }

  mi->mg = mg;
//  if(mg->iif != ifp->ifindex) VIFM_SET(ifp->ifindex, mg->oif_bitmap);

  /* Setup for PIM (*,G)  */
 
  mg->flags = (MRT_WC|MRT_SPARSE); /* Set sparsemode for PIM usage */  

  if(local_in_addr(src))
    {
      printf("Report from own interface %s for group %s\n", 
	     ifp->name, inet_2a(mg->group, b1));
      mg->flags |= MRT_LOCAL;
      mi->flags |= MRT_LOCAL;
    }
  else
    {
      mg->flags |= MRT_CONNECTED;
      mi->flags |= MRT_CONNECTED;
    }
  mg->flags |= MRT_IGMP_IN_USE;
  mi->flags |= MRT_IGMP_IN_USE;

  ZBUG_G(mg->group)
    {
      fprintf(ZBUG_LOG, "Added IGMP oif %s ", ifp->name);
      pim_flags_f(ZBUG_LOG, 0, mg->group, mi->flags);
      fprintf(ZBUG_LOG, "Group stats ");
      pim_flags_f(ZBUG_LOG, 0, mg->group, mi->flags);
      fflush(ZBUG_LOG);
    }

  if(mg->flags & MRT_BLACKHOLE) 
    {
      mg->flags &= ~MRT_BLACKHOLE; 
      sync_mg_family(mg);
      ZBUG_G(mg->group)
	{
	  fprintf(ZBUG_LOG, "Blackhole -> IGMP ");
	  pim_flags_f(ZBUG_LOG, 0, mg->group, mg->flags);
	  fflush(ZBUG_LOG);
	}
    }  
  
  {
    uint32_t rp = select_rp( group );
    struct msrc *ms;
    
    ms = ms_lookup(rp);

    if(!ms)
      {
	ms = ms_add(rp);
	ms->use_count++;
      }

    mg->pim.rp = ms;
  }
  

  /* Update oifs for all (S,G) in this (*,G) */

  add_oif_to_mr(mg, mi->ifp);

  if(mg->pim.rp->ifp)
    trigger_join_prune(mg->pim.rp->ifp);

  return 0;
}

int
parse_igmp_report(struct in_addr *src, struct igmphdr *igmp, 
		 struct interface *ifp)
{
  struct pim_interface *pi;
  zassert(ifp->info, -1);
  pi = ifp->info;

  process_igmp_report(src, igmp->group, ifp, 0);
  return 0;
}

void
init_igmp_timers(struct pim_interface *pi)
{

/* Timers and default values from RFC 2236 */

  pi->Robustness_Variable = 2; /* Default 2 */
  pi->Query_Interval = 125; /* 125 s */
  pi->Query_Response_Interval = 100; /* 10 s */
  pi->Group_Membership_Interval = pi->Robustness_Variable*pi->Query_Interval 
    + pi->Query_Response_Interval;
  pi->Other_Querier_Present_Interval = 
    pi->Robustness_Variable*pi->Query_Interval + 
    pi->Query_Response_Interval/2;

  pi->Startup_Query_Interval = pi->Query_Interval/4;
  pi->Startup_Query_Count = pi->Robustness_Variable;
  pi->Last_Member_Query_Interval = 10; /* 1 s */
  pi->Last_Member_Query_Count =  pi->Robustness_Variable;
  pi->Unsolicited_Report_Interval = 10; /* 10 s */
  pi->Version_1_Router_Present_Timeout = 400; /* 400 s */
}

void
parse_igmp_packet(char *p, int len, struct interface *ifp)
{
  struct ip *ip;
  struct igmphdr *igmp;
  struct in_addr src;
  int ip_hlen, ip_len;
  
  ip = (struct ip *)p ;
  ip_hlen = ip->ip_hl*4; 
  ip_len = ntohs(ip->ip_len);

  
  src = ip->ip_src;

  /* TODO: CHECKSUM TEST & PROTOCOL TEST*/
  /* TODO: Check IP options. Router Alert*/  

  if(ip->ip_p == 0) {
    ZBUG(ZBUG_PIM_KERNEL_PACKET)
      zlog_warn("ip_len %d, proto = %d\n", ip_len, ip->ip_p);
    process_kernelCall(p, ifp);
    return;
  }
    igmp = (struct igmphdr *) (p+ip_hlen);

    switch(igmp->type) {

      /* Parse all types from igmp.h here */
      
    case IGMP_HOST_MEMBERSHIP_QUERY:
      parse_igmp_query(&src, igmp, ifp);
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_HOST_MEMBERSHIP_QUERY from %s\n", inet_ntoa(src));
    break;
      
    case IGMP_HOST_MEMBERSHIP_REPORT:
      parse_igmp_report(&src, igmp, ifp);
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_HOST_MEMBERSHIP_REPORT from %s\n", inet_ntoa(src));
    break;

    case IGMP_DVMRP:
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_DVMRP from %s\n", inet_ntoa(src));
    break;

    case IGMP_PIM:
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_PIM from %s\n", inet_ntoa(src));
    break;

    case IGMP_TRACE:
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_TRACE from %s\n", inet_ntoa(src));
    break;

    case IGMP_HOST_NEW_MEMBERSHIP_REPORT:
      parse_igmp_report(&src, igmp, ifp);
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_HOST_NEW_MEMBERSHIP_REPORT from %s\n", inet_ntoa(src));
    break;

    case IGMP_HOST_LEAVE_MESSAGE:
      parse_igmp_leave(&src, igmp, ifp);
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_IGMP_HOST_LEAVE_MESSAGE from %s\n", inet_ntoa(src));
    break;
    
    case IGMP_MTRACE_RESP:
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_MTRACE_RESP from %s\n", inet_ntoa(src));
    break;

    case IGMP_MTRACE:
    ZBUG(ZBUG_IGMP_MESSAGES) 
      printf("RX IGMP_MTRACE from %s\n", inet_ntoa(src));
    break;
    
    default:
      
    }
}

int igmp_read_raw(struct thread *r)
{
  struct interface *ifp;
  char buf[1500];
  int ret, ifindex;
  
  int sock = THREAD_FD (r);
  t_igmp_raw = thread_add_read (master, igmp_read_raw, NULL, sock);
  
  ret = pim_recvmsg (sock, buf, 1500, &ifindex);
  
  if (ret < 0) printf ("Error = %d", ret);
  else {
    int i;
    ZBUG(ZBUG_PIM_PACKET) 
      {
	ZBUGF("RX IGMP (idx %d) ", ifindex);
	for(i=0; i < ret; i++) ZBUGF ("%x ", buf[i]&0xFF);
	ZBUGF("\n"); 
      }
  }
  ifp = get_iflist_ifp(ifindex);
  if( ifp == NULL) return 0;
  parse_igmp_packet(buf, ret, ifp);
  return ret;
}

void
igmp_query_neighbors(struct interface *ifp, u_int32_t group, 
		     u_int32_t interval, u_int32_t dest)
{
  struct igmphdr *igmp;
  u_int32_t ip_opt;

  igmp = (struct igmphdr *) (igmp_sendbuf + sizeof(struct ip) + 
			     +sizeof(ip_opt));

  igmp->type  = IGMP_HOST_MEMBERSHIP_QUERY;
  igmp->code  = interval;
  igmp->csum  = 0;
  igmp->group = group;

  /* Set Router Alert IP Option */
  ip_opt = ((u_int32_t) (0x94)<<24) + ((u_int32_t) (0x4)<<16); 
  
  send_igmp_packet( ifp, igmp_sendbuf, sizeof(struct igmphdr), 
		    dest, ip_opt);
}

void
send_igmp_packet(struct interface *ifp, char *buf, int len, u_int32_t addr,
		 u_int32_t ip_opt)
{
 static struct sockaddr_in sockdst = {AF_INET};
 struct pim_interface *pi;
 struct ip *ip;
 struct igmphdr *igmp;
 struct msghdr *msg;
 struct cmsghdr *cmsg;
 struct iovec iovector;
 char msgbuf[256];
 struct in_pktinfo *pktinfo;
 u_long src, dst;
 char b;

 if (!ifp->flags & IFF_UP) return;

 src = htonl(0);
 dst = htonl(addr);
 ip = (struct ip *) buf;
 ip->ip_hl = sizeof(struct ip) >> 2;
  ip->ip_v = IPVERSION;
 ip->ip_tos = 0xC0;
 ip->ip_off = 0L;
 ip->ip_p = IPPROTO_IGMP;
 ip->ip_ttl = 1;

 ip->ip_src.s_addr = src;
 ip->ip_dst.s_addr = dst;

 if(ip_opt) {
   ip_opt = htonl(ip_opt);
   memcpy(&buf[ip->ip_hl*4], &ip_opt, 4);
   ip->ip_hl++;
 }

 igmp = (struct igmphdr *) (buf + ip->ip_hl*4);
 pi=ifp->info;
 ip->ip_len  = ip->ip_hl*4  + len;

 b = 0; 
 if( setsockopt(igmp_sock, IPPROTO_IP, IP_MULTICAST_LOOP, 
		  (char *)&b,sizeof(b)) < 0)
   zlog_warn("sendto %s", strerror (errno));
 
 b = 1;
 if (setsockopt(igmp_sock, IPPROTO_IP, IP_HDRINCL,
		(char *) &b, sizeof(b)) < 0)
   zlog_warn("sendto %s", strerror (errno));

 b = 0;
 if (setsockopt(igmp_sock, IPPROTO_IP, SO_BROADCAST,
		(char *) &b, sizeof(b)) < 0)
   zlog_warn("sendto %s", strerror (errno));
 
 igmp->csum = in_cksum((u_short *) igmp, sizeof(struct igmphdr)+len);

 bzero(&sockdst,sizeof(sockdst));
 sockdst.sin_family=AF_INET;
 sockdst.sin_addr.s_addr = dst;

 cmsg = (struct cmsghdr *) (msgbuf + sizeof(struct msghdr));
 cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(struct in_pktinfo);
 cmsg->cmsg_level = SOL_IP;
 cmsg->cmsg_type = IP_PKTINFO;
 pktinfo = (struct in_pktinfo *) CMSG_DATA(cmsg);
 pktinfo->ipi_ifindex = ifp->ifindex;
 pktinfo->ipi_spec_dst.s_addr = src;
 pktinfo->ipi_addr.s_addr = src;
 
 iovector.iov_base = (void *) buf;
 iovector.iov_len = ip->ip_len; 
 msg = (struct msghdr *) msgbuf;
 msg->msg_name = &sockdst;
 msg->msg_namelen = sizeof(sockdst);
 msg->msg_iov = &iovector;
 msg->msg_iovlen = 1;
 msg->msg_control = cmsg;
 msg->msg_controllen = cmsg->cmsg_len;
 
 if (sendmsg(igmp_sock, msg, 0) < 0) {
   zlog_warn("sendto %s", strerror (errno));
 }
 /*  printf("TX on %s idx %d\n", ifp->name, ifp->ifindex); */
}

DEFUN (show_ip_igmp_interface,
       show_ip_igmp_interface_cmd,
       "show ip igmp interface",
       "show ip IGMP interface info\n"
       "View igmp interface info\n")
{
  struct interface *ifp;
  struct pim_interface *pi;
  listnode node;

  for (node = listhead (iflist); node; nextnode (node))
    {
      ifp = getdata (node);
      pi = ifp->info;

      vty_out(vty, "Interface %s %s", ifp->name, VTY_NEWLINE); 

      vty_out(vty, " IGMP running version is %s%s", 
	      pi->flags&IF_IGMP_V1? "V1":"V2",  VTY_NEWLINE);
      
      vty_out(vty, " IGMP querier %s%s", 
	      pi->flags&IF_IGMP_QUERIER? "Yes":"No",  VTY_NEWLINE);
      
      if(pi->flags&IF_DR) 
      vty_out(vty, " Interface is DR%s",  VTY_NEWLINE);

      vty_out(vty, " IGMP robustness variable %d %s",  
	      pi->Robustness_Variable, VTY_NEWLINE);

      vty_out(vty, " IGMP query interval is %d seconds %s",  
	      pi->Query_Interval, VTY_NEWLINE);

      vty_out(vty, " IGMP other querier timeout is %d seconds %s",  
	      pi->Other_Querier_Present_Interval, VTY_NEWLINE);

      vty_out(vty, " IGMP max query response time is %d seconds %s",  
	      pi->Query_Response_Interval/10, VTY_NEWLINE);

      if(pi->igmp_leave_process)
      vty_out(vty, " IGMP leave process is active %s",  VTY_NEWLINE);


    }
  return 1;
}

int mgrp_list(struct vty *vty)
{
  struct mgrp *g;
  struct oif *mi;
  listnode node, node1;
  int i;
  
  for(i=0;i <  MGRP_HASH_SIZE; i++) {
    if( mgrp_hash[i] == 0)  continue;
    for(node1=listhead (mgrp_hash[i]); node1; nextnode (node1)) {
      g = getdata(node1);
//      if(g) {
	
	for (node = listhead (g->oif_list); node; nextnode (node)) {
	  mi = getdata (node);

	  if(!mi->flags & MRT_IGMP_IN_USE) continue;

	  vty_out(vty,"%-16s", inet_2a(g->group, b1));
	  vty_out(vty, " ");

	  uptime_vty (vty, &mi->uptime);
	  vty_out(vty, "  ");
	  exptime_vty (vty, &mi->igmp.exptime_igmp);
	  vty_out(vty, "     %-16s", inet_2a(mi->igmp.last_reporter, b1));

	  vty_out(vty, "  %-8s", mi->ifp->name);
	  vty_out(vty, "%s", VTY_NEWLINE);
//	}
      }
    }
  }
  return 0;
}


DEFUN (show_ip_igmp_groups,
       show_ip_igmp_groups_cmd,
       "show ip igmp groups",
       "show ip igmp groups\n"
       "View IGMP groups\n")
{

  vty_out(vty, "IGMP connected groups%s", VTY_NEWLINE);
  vty_out(vty, "Group Address    Uptime    Expires      Last Reporter");
  vty_out(vty, "     Interface");
  vty_out(vty, "%s", VTY_NEWLINE);

  mgrp_list(vty);
  return 1;
}


DEFUN (if_ip_igmp_version,
       if_ip_igmp_version_cmd,
       "ip igmp version 1",
       "ip igmp version 1\n"
       "Specify IGMP version 1\n")
{
  struct interface *ifp;
  struct pim_interface *pi;

  ifp = vty->index;
  pi = ifp->info;
  pi->igmp_version_conf = 1;
  return CMD_SUCCESS;
}

DEFUN (if_no_ip_igmp_version,
       if_no_ip_igmp_version_cmd,
       "no ip igmp version 1",
       NO_STR
       "IGMP interface commands\n"
       "Specify IGMP version 1\n")
{
  struct interface *ifp;
  struct pim_interface *pi;

  ifp = vty->index;
  pi = ifp->info;
  pi->igmp_version_conf = 2; /* Default */
  return CMD_SUCCESS;
}

DEFUN (if_ip_igmp_join_group,
       if_ip_igmp_join_group_cmd,
       "ip igmp join-group A.B.C.D",
       "Set igmp group\n"
       "Add group permanent\n")
{
  struct interface *ifp;
  struct pim_interface *pi;
  struct in_addr group;
  listnode n;
  u_int32_t *p;
  int ret;

  ifp = vty->index;
  pi = ifp->info;

  ret = inet_aton (argv[0], &group);
  if (!ret)
    {
      vty_out (vty, "Malformed igmp address %s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  if(!IN_MULTICAST(htonl(group.s_addr))) {
    vty_out (vty, "Not a multicast address %s", VTY_NEWLINE);
    return CMD_WARNING;
  }

  /* TODO: A better handling of not installed interfaces */

  if(!pi->igmp_static_group_list) {
    vty_out (vty, "Interface not active %s", VTY_NEWLINE);
	return CMD_SUCCESS;
  }
  
  if(pi->igmp_static_group_list) {
    for (n = listhead (pi->igmp_static_group_list); n; nextnode (n)) {
      u_int32_t *g  = (u_int32_t *) getdata (n);
      if(group.s_addr == *g) {
	vty_out (vty, "Group is already joined %s", VTY_NEWLINE);
	return CMD_SUCCESS;
      }
    }
  }

   p = XMALLOC (MTYPE_PIM_OIF, sizeof (u_int32_t)); 
   *p = group.s_addr;
   
   list_add_node(pi->igmp_static_group_list, p);
   process_igmp_report(&pi->address->u.prefix4, *p, ifp, 1);

  return CMD_SUCCESS;
}

DEFUN (if_no_ip_igmp_join_group,
       if_no_ip_igmp_join_group_cmd,
       "no ip igmp join-group A.B.C.D",
       "Unset igmp group\n"
       "Remove permanent group\n")
{
  struct interface *ifp;
  struct pim_interface *pi;
  struct in_addr group;
  listnode n;
  int ret;

  ifp = vty->index;
  pi = ifp->info;

  ret = inet_aton (argv[0], &group);
  if (!ret) {
    vty_out (vty, "Malformed igmp address %s", VTY_NEWLINE);
    return CMD_WARNING;
  }

  if(pi->igmp_static_group_list) {
    for (n = listhead (pi->igmp_static_group_list); n; nextnode (n)) {
      u_int32_t *g  = (u_int32_t *) getdata (n);
      if(group.s_addr == *g) {
	list_delete_node(pi->igmp_static_group_list, n);
	free(g); /* Should not be freed with XFREE */
	return CMD_SUCCESS;
      }
    }
  }
  return CMD_SUCCESS;
}
void
igmp_interface_init()
{
 install_element (VIEW_NODE, &show_ip_igmp_interface_cmd);
 install_element (VIEW_NODE, &show_ip_igmp_groups_cmd); 

 install_element (ENABLE_NODE, &show_ip_igmp_interface_cmd);
 install_element (ENABLE_NODE, &show_ip_igmp_groups_cmd);

 install_element (INTERFACE_NODE, &if_ip_igmp_version_cmd);
 install_element (INTERFACE_NODE, &if_no_ip_igmp_version_cmd);
 install_element (INTERFACE_NODE, &if_ip_igmp_join_group_cmd);
 install_element (INTERFACE_NODE, &if_no_ip_igmp_join_group_cmd);
}
