/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>
#include "vector.h"
#include "vty.h"
#include "command.h"
#include "getopt.h"
#include "thread.h"

#include "stream.h"
#include "network.h"
#include "prefix.h"
#include "version.h"
#include "memory.h"

#include "if.h"
#include "sockunion.h"

#include "log.h"
#include "pim.h"
#include "pimd.h"
#include "pim_fsm.h"

#include <netinet/ip_mroute.h>
#include "tables.h"
#include "pim_route.h"

#include <netinet/igmp.h>
#include "if_ioctl.h"
#include "pim_packet.h"
#include "igmp_def.h"

#include "zbug.h"
ZBUG_MAIN

/* pimd options, we use GNU getopt library. */
struct option longopts[] = 
{
  { "daemon",      no_argument,       NULL, 'd'},
  { "config_file", required_argument, NULL, 'f'},
  { "vty_port",    required_argument, NULL, 'P'},
  { "version",     no_argument,       NULL, 'v'},
  { "help",        no_argument,       NULL, 'h'},
  { 0 }
};

/* Configuration file and directory. */
char config_current[] = PIM_DEFAULT_CONFIG;
char config_default[] = SYSCONFDIR PIM_DEFAULT_CONFIG;

/* pimd program name. */
char *progname;

/* Route retain mode flag. */
int retain_mode = 0;

/* Master of threads. */
struct thread_master *master;
struct thread *t_pim_raw;
struct thread *t_igmp_raw;
struct thread *t_pim_data;

/* Lists */
list pim_neighbor_list;

char *config_file = NULL;
int vty_port = PIM_VTY_PORT;

char b1[16], b2[16], b3[16], b4[16];  /* For inet_2a */
extern u_int32_t rp;

static void
usage (int status)
{
  if (status != 0)
    fprintf (stderr, "Try `%s --help' for more information.\n", progname);
  else
    {    
      printf ("Usage : %s [OPTION...]\n\n\
Daemon which manages kernel routing table management and \
redistribution between different routing protocols.\n\n\
-d, --daemon       Runs in daemon mode\n\
-f, --config_file  Set configuration file name\n\
-P, --vty_port     Set vty's port number\n\
-v, --version      Print program version\n\
-h, --help         Display this help and exit\n\
\n\
Report bugs to %s\n", progname, ZEBRA_BUG_ADDRESS);
    }

  exit (status);
}

/* SIGHUP handler. */
void 
sighup (int sig)
{
  zlog (NULL, LOG_INFO, "SIGHUP received");

  /* Terminate all thread. */
  pim_terminate ();
  pim_reset ();
  zlog_info ("pimd restarting!");

  /* Reload config file. */
  vty_read_config (config_file, config_current, config_default);

  /* Create VTY's socket */
  vty_serv_sock (vty_port ? vty_port : PIM_VTY_PORT, PIM_VTYSH_PATH); 

  /* Try to return to normal operation. */
}

/* SIGINT handler. */
void
sigint (int sig)
{
  zlog (NULL, LOG_INFO, "Terminating on signal");

  if (!retain_mode)
    pim_terminate ();

  exit (0);
}

/* SIGUSR1 handler. */
void
sigusr1 (int sig)
{
  zlog_rotate (NULL);
}

/* Signale wrapper. */
RETSIGTYPE *
signal_set (int signo, void (*func)(int))
{
  int ret;
  struct sigaction sig;
  struct sigaction osig;

  sig.sa_handler = func;
  sigemptyset (&sig.sa_mask);
  sig.sa_flags = 0;
#ifdef SA_RESTART
  sig.sa_flags |= SA_RESTART;
#endif /* SA_RESTART */

  ret = sigaction (signo, &sig, &osig);

  if (ret < 0) 
    return (SIG_ERR);
  else
    return (osig.sa_handler);
}

/* Initialization of signal handles. */
void
signal_init ()
{
  signal_set (SIGHUP, sighup);
  signal_set (SIGINT, sigint);
  signal_set (SIGTERM, sigint);
  signal_set (SIGPIPE, SIG_IGN);
  signal_set (SIGUSR1, sigusr1);
}

int
hello_thread(struct thread *t_hello)
{
  u_int32_t hello_holdtime;
  struct pim_interface *pi = THREAD_ARG (t_hello);
  struct interface *ifp;
  ifp = pi->ifp;

  /*   printf("Hello %p %p %s\n", t_hello, pi, ifp->name); */

  hello_holdtime = 3*(pi->hello_period) + ((pi->hello_period)>>1);

  pi->hello_sent++;
  send_pim_hello(ifp, hello_holdtime);

  t_hello = thread_add_timer(master, hello_thread, pi, pi->hello_period);
  pi->t_hello = t_hello;
  return 0;
}

/* Main routine of pimd. Treatment of argument and start pim finite
   state machine is handled at here. */

int
main (int argc, char **argv)
{
  char *p;
  int opt;
  int daemon_mode = 0;

  struct thread thread;

  /* Preserve name of myself. */
  progname = ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);

  zlog_default = openzlog (progname, ZLOG_NOLOG, ZLOG_PIM,
                           LOG_CONS|LOG_NDELAY|LOG_PID, LOG_DAEMON);

  /* Command line argument treatment. */
  while (1) 
    {
      opt = getopt_long (argc, argv, "df:hp:P:rv", longopts, 0);
    
      if (opt == EOF)
        break;

      switch (opt) 
        {
        case 0:
          break;
        case 'd':
          daemon_mode = 1;
          break;
        case 'f':
          config_file = optarg;
          break;
        case 'P':
          vty_port = atoi (optarg);
          break;
        case 'v':
          print_version ();
          exit (0);
          break;
        case 'h':
          usage (0);
          break;
        default:
          usage (1);
          break;
        }
    }

 /* Make thread master. */
  master = thread_make_master ();

  /* Initializations. */
  tables_init();
  signal_init ();
  debug_init();
  if_init();
  cmd_init();
  vty_init();
  pim_neighbor_list = list_init ();
  pim_init();
  memory_init();
  socket_init();
  pim_interface_init();
  igmp_interface_init();
  vifs_init ();
  kernel_init();
  rpf_init();
  pim_rp_init();
  init_pim_register_dev();
  pim_assert_init();
  data_timer_init();

/* Get configuration file. */
vty_read_config (config_file, config_current, config_default);

/* Change to the daemon program. */
  if (daemon_mode)
    daemon (0, 0);

  /* Pid file create. */
  pid_output (PATH_PIMD_PID);

  /* Create pim server */
  vty_serv_sock (vty_port ? vty_port : PIM_VTY_PORT, PIM_VTYSH_PATH);
 
  /* Execute each thread. */
  while (thread_fetch (master, &thread))
    thread_call (&thread);
    
  /* Not reached. */
  exit (0);
}

void pim_if_start(struct interface *ifp)
{
  struct pim_interface *pi;
  u_int32_t timer;

  pi = ifp->info;
  
  if (pi == NULL) {
      zlog_warn ("Interface %s structure is NULL", ifp->name);
      return;
  }

  if( ! if_get_index (ifp) ) {
    zlog_warn("Interface not in configuration %s\n", ifp->name);
    return;
  }

  if (pi->running) {
    zlog_warn("Interface is already active %s\n", ifp->name);
    return;
  }
  pi->running = TRUE;

  if_get_flags (ifp);

  if (!ifp->flags & IFF_UP) {
    zlog_warn("Interface is down %s\n", ifp->name);
  }

  if (!ifp->flags & (IFF_MULTICAST|IFF_ALLMULTI)) {
    zlog_warn("Interface is not multicast %s\n", ifp->name);
  }

  pi->address = (struct prefix *) prefix_ipv4_new ();
  pi->address->family = AF_INET;
  /*  
  pi->address->u.prefix4.s_addr = 0;
  pi->address->prefixlen = 0;
  */

  /* Get interface address & prefix */
  if_addr_ioctl (ifp, NULL);

  printf("START %s idx =%d %s/%d\n", ifp->name, ifp->ifindex,
	 inet_ntoa(pi->address->u.prefix4) ,pi->address->prefixlen );

  if_add_groups(ifp);
  init_pim_if(pi);


  /* We are both DR & querier util someone else show up */
  pi->flags =  IF_DR;
  pi->flags |= IF_IGMP_QUERIER;

  /* Start HELLO's  */
  pi->hello_sent = 0;
  pi->hello_period = HELLO_INTERVAL;
  
  /* According to the spec the first HELLO should randomized
     between 1 and HELLO_INTERVAL */
  
  timer =  (random () % HELLO_INTERVAL) + 1;
  pi->t_hello = thread_add_timer(master, hello_thread, pi, timer);


  /* It is not in specs but radomize periodic join/prunes as well 
     triggered join_prunes cover this up */

  timer =  (random () % pi->join_prune_query_interval) + 1;
  pi->t_send_join_prune = thread_add_timer(master, 
		    send_join_prune_thread, pi, timer);

  vif_add(ifp, NULL, NULL);

  igmp_if_start(ifp);
}

void pim_if_stop(struct interface *ifp)
{
  struct pim_interface *pi = ifp->info;
  
  if (pi == NULL) {
      zlog_warn ("Interface %s structure is NULL", ifp->name);
      return;
  }
  if( ! if_get_index (ifp) ) {
    zlog_warn("Interface not in configuration %s\n", ifp->name);
    return;
  }

  if (! pi->running) {
    zlog_warn("Interface is not active %s\n", ifp->name);
    return;
  }

  /* Remove all neighbors connected to this interface */
  pim_all_neighbor_remove(ifp);

  /* Stop HELLO's */
  if(pi->t_hello)  thread_cancel(pi->t_hello);
  pi->t_hello = NULL;

  /* Stop periodic sedning of join_punes */
  if(pi->t_send_join_prune)  thread_cancel(pi->t_send_join_prune);
  pi->t_send_join_prune = NULL;

  send_pim_hello(ifp, 0);
  igmp_if_stop(ifp);
  if_drop_groups(ifp);
  pi->running = FALSE;
  prefix_ipv4_free (pi->address);
  printf("STOP %s\n", ifp->name);
}

void pim_reset()
{
}

void pim_terminate()
{
}

/* PIM node structure. */
struct cmd_node pim_node =
{
  PIM_NODE,
  "%s(config-router)# ",
};

/* PIM configuration write function. */

int
config_write_pim (struct vty *vty)
{
  int write = 0;
  
  /* Router PIM statement. */
  vty_out (vty, "ip multicast-routing%s", VTY_NEWLINE);

  /* Any rp set ?*/
  if(rp)   vty_out (vty, "ip pim rp %s%s", 
		    inet_2a(rp, b1), VTY_NEWLINE);

  write++;
  return write;
}

DEFUN (debug_group,
       debug_group_cmd,
       "debug group A.B.C.D",
       "Enable debugging on group\n"
       "Enable debugging on group\n")
{
  
  /*  */
  u_int32_t g;

  if(zbug_f) {
    vty_out(vty, "Group debugging is already active %s",  VTY_NEWLINE);
    return CMD_SUCCESS;
  }
    
  g = inet_addr(argv[0]);
  if(g == -1) return CMD_WARNING;
   if(IN_MULTICAST(g))  {
     ZBUG_LOG_GROUP( g );      
     vty_out(vty, "Debug info to %s for group %s%s", 
	     ZBUG_LOGFILE,
	     inet_2a(g, b1), 
	     VTY_NEWLINE);
     fprintf(ZBUG_LOG, "Log start group %s\n", inet_2a(zbug_g, b1));
     fflush(ZBUG_LOG);
     return CMD_SUCCESS;
   }
  vty_out(vty, "Illegal address %s",  VTY_NEWLINE);
  return CMD_SUCCESS;
} 

DEFUN (no_debug_group,
       no_debug_group_cmd,
       "no debug group",
       "Disable debugging on group\n"
       "Disable debugging on group\n")
{
  
  /*  */

  ZBUG_CLOSE;
  
  vty_out(vty, "Group debugging is turned off %s",  VTY_NEWLINE);
  return CMD_SUCCESS;
} 

DEFUN (debug,
       debug_cmd,
       "debug TYPE",
       "Enable debugging\n"
       "Enable debugging\n")
{

  /*  */
   int type;

   ZBUG_TYPEOF( type, argv[0]);
   if(type)
     {
        ZBUG( type )
          {
             ZBUG_OFF_RANGE( type );
             vty_out(vty, "Debuginfo on type %s disabled%s", 
                     argv[0], 
                     VTY_NEWLINE);
          }
        else
          {
             ZBUG_ON_RANGE( type );
             vty_out(vty, "Debuginfo on type %s enabled%s", 
                     argv[0], 
                     VTY_NEWLINE);
          }
     }

  return CMD_SUCCESS;
}

DEFUN (no_debug,
       no_debug_cmd,
       "no debug",
       "Disable all debugging\n"
       "Disable all debugging\n")
{

  /*  */

   zdebug_s = zdebug = 0;
   vty_out(vty, "All debuginfo disabled%s", VTY_NEWLINE);

  return CMD_SUCCESS;
}



DEFUN (show_debug_types,
       show_debug_types_cmd,
       "show debug types",
       "Show available debug types\n"
       "Show available debug types\n")
{

  /*  */
   int i;
   if(zdebug_array)
     for( i = 1; i<MAX_ZBUG; i++ )
       if(zdebug_array[i])
	 vty_out(vty, "\t%s%s", zdebug_array[i], VTY_NEWLINE);
   
   return CMD_SUCCESS;
}

DEFUN (show_debug_status,
              show_debug_status_cmd,
              "show debug status",
              "Show enabled debug types\n"
              "Show enabled debug types\n")
{
   /*  */
   int i;
   
   vty_out(vty, "Currently enabled debug types:%s", VTY_NEWLINE);
   if(zdebug_array)
     for( i = 1; i<MAX_ZBUG; i++ )
       if(zdebug_array[i])
     {	
	ZBUG( i )
	  vty_out(vty, "\t%s%s", zdebug_array[i], VTY_NEWLINE);
     }  
   return CMD_SUCCESS;
}

DEFUN (ip_multicast_routing,
       ip_multicast_routing_cmd,
       "ip multicast-routing",
       "Enable PIM multicast routing process\n"
       "Enable PIM multicast\n")
{

  /* Change to PIM node. */
    vty->node = PIM_NODE;

  return CMD_SUCCESS;
}

DEFUN (no_ip_multicast_routing,
       no_ip_multicast_routing_cmd,
       "no ip multicast-routing",
       "Disable PIM multicast routing process\n"
       "Disable PIM multicast\n")
{

  /* Change to PIM node. */
    vty->node = PIM_NODE;

  return CMD_SUCCESS;
}

void
uptime_vty (struct vty *vty, time_t *t)
{

#define TIME_BUF 25
#define ONE_DAY_SECOND 60*60*24
#define ONE_WEEK_SECOND 60*60*24*7

  time_t uptime;
  struct tm *tm;
  char timebuf [TIME_BUF];

  /* If there is no connection has been done before print `never'. */
  if (*t == 0)
    {
      vty_out (vty, "never   ");
      return;
    }

  /* Get current time. */
  uptime = time (NULL);
  uptime -= *t;
  tm = gmtime (&uptime);

  /* Making formatted timer strings. */
  if (uptime < ONE_DAY_SECOND)
    snprintf (timebuf, TIME_BUF, "%02d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->
tm_sec);
  else if (uptime < ONE_WEEK_SECOND)
    snprintf (timebuf, TIME_BUF, "%dd%02dh%02dm", tm->tm_yday, tm->tm_hour, tm->
tm_min);
  else
    snprintf (timebuf, TIME_BUF, "%02dw%dd%02dh", 
              tm->tm_yday/7, tm->tm_yday - ((tm->tm_yday/7) * 7), tm->tm_hour);

  /* Out puts to vty. */
  vty_out (vty, "%8s", timebuf);
}

void
exptime_vty (struct vty *vty, time_t *t)
{

#define TIME_BUF 25
#define ONE_DAY_SECOND 60*60*24
#define ONE_WEEK_SECOND 60*60*24*7

  time_t exptime;
  struct tm *tm;
  char timebuf [TIME_BUF];

  /* If there is no connection has been done before print `never'. */
  if (*t == 0)
    {
      vty_out (vty, "expired   ");
      return;
    }

  exptime = time (NULL);
  exptime = *t - exptime;;
  
  tm = gmtime (&exptime);

  /* Making formatted timer strings. */
  if (exptime < ONE_DAY_SECOND)
    snprintf (timebuf, TIME_BUF, "%02d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->
tm_sec);
  else if (exptime < ONE_WEEK_SECOND)
    snprintf (timebuf, TIME_BUF, "%dd%02dh%02dm", tm->tm_yday, tm->tm_hour, tm->
tm_min);
  else
    snprintf (timebuf, TIME_BUF, "%02dw%dd%02dh", 
              tm->tm_yday/7, tm->tm_yday - ((tm->tm_yday/7) * 7), tm->tm_hour);

  /* Out puts to vty. */
  vty_out (vty, "%8s", timebuf);
}

DEFUN (show_ip_pim_neighbor,
       show_ip_pim_neighbor_cmd,
       "show ip pim neighbor",
       "show pim neigbors\n"
       "View pim neighbor\n")
{
  struct pim_interface *pi;
  struct pim_neighbor *pn;
  listnode node;
  
  vty_out(vty, "PIM Neighbor Table%s", VTY_NEWLINE);
  vty_out(vty, 
    "Neighbor Address  Interface   Uptime    Expires      Ver  Mode%s",
	  VTY_NEWLINE);
  
  for (node = listhead (pim_neighbor_list); node; nextnode (node)) {
    pn = getdata (node);

    vty_out(vty, "%-15s ", inet_ntoa(pn->address));

    vty_out(vty,  "%6s ", pn->ifp->name);
    vty_out(vty,  "       ");
    uptime_vty (vty, &pn->uptime);
    vty_out(vty,  "  ");
    exptime_vty (vty, &pn->exptime);
    vty_out(vty,  "   ");

    vty_out(vty,  "%3d",  pn->version);

    vty_out(vty,  "          "); /* MODE */

    pi = pn->ifp->info;
    if(pi->flags & IF_DR) vty_out(vty, "(DR)");

    vty_out(vty, "%s",  VTY_NEWLINE);

  }
  return CMD_SUCCESS;
}


void
pim_init()
{

  /* Install top nodes. */

   install_element (VIEW_NODE, &show_ip_pim_neighbor_cmd);
   install_element (ENABLE_NODE, &show_ip_pim_neighbor_cmd);
   install_element (ENABLE_NODE, &debug_cmd);
   install_element (ENABLE_NODE, &debug_group_cmd);
   install_element (ENABLE_NODE, &no_debug_group_cmd);
   install_element (ENABLE_NODE, &no_debug_cmd);
   install_element (ENABLE_NODE, &show_debug_types_cmd);
   install_element (ENABLE_NODE, &show_debug_status_cmd);
   install_element (CONFIG_NODE, &ip_multicast_routing_cmd);
   install_element (CONFIG_NODE, &no_ip_multicast_routing_cmd);
   install_element (CONFIG_NODE, &debug_cmd);
   install_element (CONFIG_NODE, &debug_group_cmd);



  install_node (&pim_node, config_write_pim);
  install_element (PIM_NODE, &config_end_cmd);
  install_element (PIM_NODE, &config_exit_cmd);
  install_element (PIM_NODE, &config_help_cmd);

}

void
debug_init()
{
   ZBUG_INIT();

   ZBUG_DEF( ZBUG_PIM_AUTORP );
   ZBUG_DEF( ZBUG_PIM_PACKET );
   ZBUG_DEF( ZBUG_PIM_PACKET_V1 );
   ZBUG_DEF( ZBUG_PIM_PACKET_V2 );
   ZBUG_DEF( ZBUG_PIM_NEIGHBOR );
   ZBUG_DEF( ZBUG_PIM_ROUTE );
   ZBUG_DEF( ZBUG_PIM_REGISTER_PACKET );
   ZBUG_DEF( ZBUG_PIM_KERNEL_PACKET );
   ZBUG_DEF( ZBUG_PIM_MESSAGES );
   ZBUG_DEF( ZBUG_PIM_TIMER );
   ZBUG_DEF( ZBUG_IGMP_MESSAGES );
   ZBUG_DEF( ZBUG_IGMP_TIMER );
}

int
pim_neighbor_list_add(struct pim_neighbor *pn)
{
  list_add_node (pim_neighbor_list, pn);
  return 1;
}

struct pim_neighbor *
pim_neighbor_alloc ()
{
  struct pim_neighbor *pn;

  /* Allocate new neighbor. */
  pn = XMALLOC (MTYPE_PIM_NEIGHBOR, sizeof (struct pim_neighbor));
  bzero (pn, sizeof (struct pim_neighbor));

return pn;
}

int 
pim_neighbor_list_delete (struct pim_neighbor *pn)
{
  struct pim_neighbor *p;
  listnode node;

  for (node = listhead (pim_neighbor_list); node; nextnode (node)) {
    p = getdata (node);
    if( pn == p) {
      list_delete_node(pim_neighbor_list, node);
      return 1;
    }
  }
  return 0;
}

struct pim_neighbor *
pim_neighbor_list_lookup_from_in_addr (struct in_addr *a)
{
  struct pim_neighbor *p;
  listnode node;

  for (node = listhead (pim_neighbor_list); node; nextnode (node)) {
      p = getdata (node);
       if ( p->address.s_addr == a->s_addr ) return p;
    }
  return NULL;
}


