/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>
#include "vector.h"
#include "vty.h"
#include "command.h"
#include "getopt.h"
#include "thread.h"
#include "stream.h"
#include "network.h"
#include "prefix.h"
#include "version.h"
#include "memory.h"
#include "if.h"
#include "sockunion.h"
#include "log.h"
#include "pim.h"
#include "pimd.h"
#include "pim_fsm.h"
#include <string.h>
#include <netinet/igmp.h>
#include <linux/mroute.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "assert.h"
#include "tables.h"
#include "pim_packet.h"
#include "igmp_def.h"
#include "if_idx_nexthop.h"
#include "pim_route.h"
#include "pim_assert.h"

#include "zbug.h"
ZBUG_EXTERN

extern int pim_sock, igmp_sock;
extern list pim_neighbor_list;
extern list mrt_hash[];
extern list mgrp_hash[];
extern list mg_list;
extern list msrc_hash[MSRC_HASH_SIZE];
extern struct best_route best_route;

char b1[16], b2[16], b3[16], b4[16];  /* For inet_2a */

void
init_pim_if(struct pim_interface *pi)
{

 pi->join_prune_query_interval = 60; /* Default interval in s*/
 pi->join_prune_holdtime = 3*pi->join_prune_query_interval + 
   pi->join_prune_query_interval/2; /* Default timeout 210 in s */

 pi->join_prune_supression_timeout = pi->join_prune_query_interval +
   pi->join_prune_query_interval/4;

 pi->data_timeout = 210; /* s */
 
 pi->trigger_join_prune_delay = 1; /* s*/
 pi->trigger_join_prune_active=0;

 pi->register_ttl = 16;  

 /* Set default Assert preference and metric if no set */

 if(!(pi->flags & IF_PREFERENCE_SET))
   pi->preference = ASSERT_DEFAULT_PREFERENCE ;

 if(!(pi->flags & IF_METRIC_SET))
   pi->metric = ASSERT_DEFAULT_METRIC  ;
}

int
same_subnet(u_int32_t s)
{
  u_int32_t idx;
  idx =  if_idx_nexthop((struct in_addr *) &s);
  if( best_route.idx < 0) return 0;
  if( best_route.gw > 0) return 0;
  return 1;
}

struct mgrp *
pim_mg_add( u_int32_t group, time_t holdtime)
{
  struct mgrp *mg;
  uint32_t rp;
  struct msrc *ms;

  mg=mg_add(group);
  zassert(mg, -1);
  
  mg->pim.exptime = time(NULL)+holdtime;
  mg->pim.uptime = time(NULL);
  
  if( holdtime) 
    mg->pim.t_entry_timeout = thread_add_timer(master, rt_1_wc_1_timeout, mg, 
					       holdtime);
  rp = select_rp(group);
  ms = ms_lookup(rp);
  if(!ms)
    {
      ms = ms_add(rp);
      ms->use_count++;
    }
  
  mg->pim.rp = ms;
  return mg;
}

void
init_pim_register_dev()
{
 struct vifctl vc;
 
 /* We ignore vif the best we can. We always ifindex from kernel
    But this special device need a little bit of special handling.
    We simply as pimreg at max vif size. MAXVIF-1 is defined in  
    PIMREG_IDX.
    */

 bzero(&vc, sizeof(struct vifctl));
 vc.vifc_vifi = PIMREG_IDX;
 vc.vifc_flags = VIFF_REGISTER;
 vc.vifc_lcl_addr.s_addr = 0;
 vc.vifc_rate_limit  = 1;
 vc.vifc_threshold  = 1;
   
   if (setsockopt(igmp_sock, IPPROTO_IP, MRT_ADD_VIF, 
		  (char *)&vc, sizeof(vc)) < 0)
     printf("Error adding PIMREG vif %s\n", strerror(errno));
 return;
}

void
process_nocache_we_are_rp()
{
}

void
process_nocache_densmode()
{
}

void
process_nocache_local_group(char *buf, struct interface *ifp)
{
 struct igmpmsg *igmpctl = (struct igmpmsg *) buf;
 struct pim_interface *pi=ifp->info;
 vifbitmap_t oifs;
 struct mgrp *mg;
 struct mrt *mr;
 struct msrc *ms;

 if(! pi->mode & 
    (PIM_SPARSE_DENSE_MODE|PIM_DENSE_SLOW_START_MODE)) {
   process_nocache_densmode();
   return ;
 }
    
   if( ifp != get_iflist_ifp(igmpctl->im_vif) ) printf("VACCKO\n");

 /* We are DR for and have packets from a local group 
    Need to create a group?
    */
    
   if((mg = mg_lookup( igmpctl->im_dst.s_addr)) == NULL ) {
     mg=pim_mg_add(igmpctl->im_dst.s_addr, 0);
   }
   mg->flags |= (MRT_SPARSE|MRT_CONNECTED|MRT_REGISTER|MRT_PIM_IN_USE);

   if( !mg->pim.rp ) {
     process_nocache_densmode();
     return;
   }
   
   if(local_in_addr((struct in_addr *)&(mg->pim.rp))) {
     process_nocache_we_are_rp();
     return;
   }
   
   /* Add source if needed */
   
   if((ms= ms_lookup( igmpctl->im_src.s_addr)) == NULL ) {
     ms = ms_add( igmpctl->im_src.s_addr);
     if(!ms)
       return;
   }

   /* Entry timer?? */
   
   /* Check (S,G) */

   mr = mr_lookup(igmpctl->im_src.s_addr, igmpctl->im_dst.s_addr);
   if(mr) printf("Warning (%s, %s) already exsists for NOCACHE\n",
		 inet_2a(igmpctl->im_src.s_addr, b1),
		 inet_2a(igmpctl->im_dst.s_addr, b2));

   mr = mr_add(ms, mg);
   mr->flags |= 
     (MRT_SPARSE|MRT_CONNECTED|MRT_REGISTER|MRT_RPT|MRT_PIM_IN_USE); 

   /* TODO should we inherit (*,G) oif lit here ? */
   copy_mg_oifs(mr, mg);

   /* Our pseudo-(S,G) shouldnt have entry timouts.
      If we set a timer on these it is purely a
      garbage collect timer and not part of the spec 
      */
   mr_entry_timer_update(mr, pi->data_timeout);
   
   /*    VIFM_CLR(igmpctl->im_vif, mr->oifs); */
   oifs = mr_oifs(mr);
   VIFM_CLR(igmpctl->im_vif, oifs);
   mfc_add(igmp_sock,
	   (u_int32_t) igmpctl->im_src.s_addr,
	   (u_int32_t) igmpctl->im_dst.s_addr,
	   mr_iif( mr ),
//igmpctl->im_vif, 
	   oifs);
   
   return;
}

void
process_whole_packet(char *buf, struct interface *ifp)
{
 struct igmpmsg *igmpctl = (struct igmpmsg *) buf; 
 struct pim_interface *pi = ifp->info;
 struct mgrp *mg;
 struct mrt *mr;
 u_int32_t len;
 struct ip *ip;
 u_int32_t reserved = 0;
 char *p;

 /* To avoid copying of the register packets. The following code
    assumes:

    sizeof(struct igmpmsg) >= sizeof(struct ip) + ev IP otions
    The execution is not interrupted so the inbuf is overwritten
    before we pass it sendmsg.

    */

 ZBUG(ZBUG_PIM_REGISTER_PACKET)
    zlog_warn ("Register from %s to %s", inet_2a(igmpctl->im_src.s_addr, b1),
	       inet_2a(igmpctl->im_dst.s_addr, b2));
 p = buf;
  
 mr = mr_lookup(igmpctl->im_src.s_addr, igmpctl->im_dst.s_addr);
 if(! mr) return;
 if(! mr->flags & MRT_REGISTER) {
   ZBUG(ZBUG_PIM_REGISTER_PACKET)
    zlog_warn ("Register dropped not MRT_REGISTER");
   return;
 }
 mg = mg_lookup(igmpctl->im_dst.s_addr);
 if(!mg) return;
 if(!mg->pim.rp) return;

 /* Update timer */
 mr_entry_timer_update(mr, pi->data_timeout);

 p += sizeof(struct igmpmsg);
 ip = (struct ip *) p;

#if 0
 ip->ip_sum = 0;
 ip->ip_sum = in_cksum((u_short *) ip, ip->ip_hl << 2);
#endif

 len = ntohs(ip->ip_len) + sizeof(struct pimheader) +4;

 p -= 4; /* Add the reserved header */

 reserved = 0;
 memcpy(p, &reserved, 4);

// RPF Neighbor

 p -= sizeof(struct pimheader);
 p -= sizeof(struct ip);


 send_pim_v2_packet(mg->pim.rp->ifp, p, len,
		    PIM_REGISTER_WHOLEPACKET_CRC, mg->pim.rp->src, 
		    pi->register_ttl);

 /*
   send_pim_v2_packet(mg->pim.rp->ifp, p, len, PIM_REGISTER, mg->pim.rp->src, 
   pi->register_ttl);
		    */		     
}

void send_pim_register(struct interface *ifp, void *buf, u_int32_t len, 
		       u_int32_t dst)
{
  struct pim_interface *pi = ifp->info;
  
   send_pim_v2_packet(pi->ifp, buf, len, PIM_REGISTER, dst, 
		      pi->register_ttl);  
}

void
process_wrong_vif(char *buf, struct interface *ifp)
{
  struct igmpmsg *igmpctl = (struct igmpmsg *) buf; 
  struct mgrp *mg;
  struct mrt *mr;
  struct oif *mi;
  vifbitmap_t oifs;
  u_int32_t iif, rpt, our_metric, our_preference;

  printf(" WRONGVIF (%s, %s) on %s\n",
	 inet_2a(igmpctl->im_src.s_addr, b1), 
	 inet_2a(igmpctl->im_dst.s_addr, b2),
	 ifp->name);

    /* We have got a packet received on wrong iif.
       This happens when we have a mrt entry for (S,G) towards RP
       and receive a packet using SPT.

      If we are on transition state in other words we have the RPT-bit 
      cleared. We should complete the transition to SPT.
   */

  mg = mg_lookup(igmpctl->im_dst.s_addr);
  if(! mg) {
    printf(" WRONGVIF and no group entry for (%s, %s)\n",
	   inet_2a(igmpctl->im_src.s_addr, b1), 
	   inet_2a(igmpctl->im_dst.s_addr, b2));
  }

  mr = mr_lookup(igmpctl->im_src.s_addr, igmpctl->im_dst.s_addr);
  if(! mr) {
    printf(" WRONGVIF and no mrt entry for (%s, %s)\n",
	   inet_2a(igmpctl->im_src.s_addr, b1), 
	   inet_2a(igmpctl->im_dst.s_addr, b2));
  }

  if(!(mg->flags & MRT_PIM_IN_USE)) return ; /* PIM active  ? */
  if(mr->flags & MRT_REGISTER) return ; /* Register ? */

  mi =mr_oif_lookup(mr, ifp);
  if(mi && ( mi->flags & MRT_ASSERTED)) return;


  iif = mr->ms->ifp->ifindex;

  /* Switch to SPT if SPT=RPT=0 and WRONGVIF on if towars source */

  if(!(mr->flags & MRT_SPT))
    if(!(mr->flags & MRT_RPT) && 
       (iif == ifp->ifindex)) {

      mr->flags |= MRT_SPT;
      mr_oif_delete(mr, mr->ms->ifp);
      mr_add_oif(mr, mg->pim.rp->ifp);    /* Thinking TODO: */
      oifs = mr_oifs(mr);
      mfc_add(igmp_sock,
	      (u_int32_t) mr->ms->src,
	      (u_int32_t) mr->mg->group,
	      iif, 
	      oifs);
      return;
  }

  if( ! mi ) return;
  if(mr->flags & MRT_BLACKHOLE) return ; 

  /* Assert time ? */
  
  if(mr->flags & MRT_SPT) rpt = 0;
  else rpt = 1;

  if(rpt) {
    our_metric = mg_get_metric(mg, ifp);
    our_preference = mg_get_preference(mg, ifp);
    
    send_pim_assert(ifp, 
		    mg->group, 
		    mr->ms->src,
		    rpt,
		    our_preference, 
		    our_metric);
  }
  else {
    our_metric = mr_get_metric(mr, ifp);
    our_preference = mr_get_preference(mr, ifp);
    send_pim_assert(ifp, 
		    mg->group, 
		    mr->ms->src,
		    rpt,
		    our_preference, 
		    our_metric);
  }
}

void
process_kernelCall(char *buf, struct interface *ifp)
{
  struct igmpmsg *igmpctl = (struct igmpmsg *) buf;
  struct pim_interface *pi = ifp->info;
  vifbitmap_t oifs;
   u_int32_t idx, iif;
  struct mgrp *mg;
  struct mrt *mr;
 struct msrc *ms;

  switch (igmpctl->im_msgtype) {  
  case IGMPMSG_NOCACHE:
    printf(" NOCACHE \n");
    oifs = 0;

    /* Cases for NOCACHE
       1. Packet via RP-tree from new source, where G matches (*,G)
       2. First packet from S matching (S,G) entry with SPT = 0
       -->Set SPT bit, mfc_add route, send prune towards RP
       3. Local source
       */

    idx =  if_idx_nexthop((struct in_addr *) &igmpctl->im_src.s_addr);
     
    /* We have local source and we are DR */
 
    if( (pi->flags & IF_DR) & (best_route.gw == 0)) {
        process_nocache_local_group(buf, ifp);
	return;
    }

    if( idx < 0) {
      printf("Error no interface for incomming source %s\n",
	     inet_2a(igmpctl->im_src.s_addr, b1 ));
    }
    
    if((mg = mg_lookup( igmpctl->im_dst.s_addr)) == NULL ) {

      /* Special case. To avoid NOCACHE for traffic destinated to 
	 a host on same LAN. Where we not have this group joined. */

      mg = pim_mg_add( igmpctl->im_dst.s_addr, pi->data_timeout);
      mg->flags |=  MRT_BLACKHOLE;
      ZBUG_G(mg->group)
	pim_flags_f(ZBUG_LOG, 0, mg->group, mg->flags);
    }
    
    printf("Process kernelCall iif=0x%x, kiif=0x%x G=%s, S=%s\n",
	   mg->pim.rp->ifp->ifindex, igmpctl->im_vif, 
	   inet_2a(igmpctl->im_dst.s_addr, b1),
	   inet_2a(igmpctl->im_src.s_addr, b2));
    
    if(( mg->pim.rp->ifp->ifindex == igmpctl->im_vif) &&
       (mg->pim.rp->ifp->ifindex >= 0)) {
      
      /* A new source? */
      
      if((ms= ms_lookup( igmpctl->im_src.s_addr)) == NULL ) {
	ms = ms_add( igmpctl->im_src.s_addr);
	if(!ms)
	  return;
      }

      /* Add (S,G) */
      
      mr = mr_add(ms, mg);
      mr->flags |= (MRT_RPT);
      copy_mg_oifs(mr, mg);

      mr_entry_timer_update(mr, pi->data_timeout);      
      
      oifs = mr_oifs(mr);
      iif = mr_iif(mr);
      VIFM_CLR(igmpctl->im_vif, oifs); /* Remove?  */
      mfc_add(igmp_sock,
	      (u_int32_t) igmpctl->im_src.s_addr,
	      (u_int32_t) igmpctl->im_dst.s_addr,
	      iif,
//	      igmpctl->im_vif, 
	      oifs);
    } else printf("NOCACHE missing block\n");
    break;
    
  case IGMPMSG_WRONGVIF:
    process_wrong_vif(buf, ifp);
    break;
    
  case IGMPMSG_WHOLEPKT:
    /* ALL multicast packets using this interface
       will be sent here.
       setsockopt() with MRT_ADD_VIF and VIFF_REGISTER
       */

    process_whole_packet(buf, ifp);
    break;

  default:
    printf(" ERROR UNSUPPORTED KERNEL MESSAGE\n");
    
  }
}

void 
mr_init()
{
  int b = 1;

  printf("igmp_sock = %x\n", igmp_sock);

  if (setsockopt(igmp_sock, IPPROTO_IP, MRT_INIT,(char *) &b, sizeof(b)) < 0)
    printf("Can't start multicast-routing. %s\n", strerror(errno));

  if (setsockopt(igmp_sock, IPPROTO_IP, MRT_PIM, (char *) &b, sizeof(b)) < 0)
    printf("Can't set PIM-routing. %s\n", strerror(errno));

  route_cmd_init();

#if 1
  /* Start timing out unused sources from hash */
  t_ms_timeout = 
    thread_add_timer(master, ms_timeout_thread, NULL, MS_PEROIDIC_TIMER ); 
#endif

}

void
kernel_init()
{
  mr_init();
}

int vif_add(struct interface *ifp, u_int32_t *flags, u_int32_t *threshold)
{
  struct pim_interface *pi = ifp->info;
  struct vifctl vc;


  bzero(&vc, sizeof(struct vifctl));
  vc.vifc_vifi          = ifp->ifindex;
  if(flags) vc.vifc_flags         = *flags;
  else vc.vifc_flags         = 0;

  if( threshold)  vc.vifc_threshold     = *threshold;
  else  vc.vifc_threshold  = 1;

  vc.vifc_lcl_addr.s_addr  =  pi->address->u.prefix4.s_addr;//  v->local;
  vc.vifc_rate_limit      = 1; //v->rate_limit;
  
  if (setsockopt(igmp_sock, IPPROTO_IP, MRT_ADD_VIF, 
		 (char *)&vc, sizeof(vc)) < 0)
    printf("Error adding vif %s\n", strerror(errno));
   return 0;
}

int
mfc_add(int socket, u_int32_t src, u_int32_t group, vifi_t iif, 
          vifbitmap_t oifs)
{
    struct mfcctl mc;
    vifi_t vifi;

    mc.mfcc_origin.s_addr = src;
    mc.mfcc_mcastgrp.s_addr = group;
    mc.mfcc_parent = iif;
    for (vifi = 0; vifi < MAXVIFS ; vifi++) {
      if (VIFM_ISSET(vifi, oifs)) {
	mc.mfcc_ttls[vifi] = 1;
      }
      else
	mc.mfcc_ttls[vifi] = 0;
    }
    if (setsockopt(socket, IPPROTO_IP, MRT_ADD_MFC, (char *)&mc,
                sizeof(mc)) < 0) {
      printf("Can't add mroute for %s\n", strerror(errno));
      return -1;
    }
    return 1;
}

int
mfc_delete(int socket, u_int32_t src, u_int32_t group, vifi_t iif, 
	  vifbitmap_t oifs)
{
    struct mfcctl mc;
    vifi_t vifi;

    mc.mfcc_origin.s_addr = src;
    mc.mfcc_mcastgrp.s_addr = group;
    mc.mfcc_parent = iif;
    for (vifi = 0; vifi < MAXVIFS; vifi++) {
        if (VIFM_ISSET(vifi, oifs)) {
	  mc.mfcc_ttls[vifi] = 1;
	  printf("DELETING vifi = %d\n", vifi);
	}
        else
            mc.mfcc_ttls[vifi] = 0;
    }
    if (setsockopt(socket, IPPROTO_IP, MRT_DEL_MFC, (char *)&mc,
		sizeof(mc)) < 0) {
      printf("Can't delete mroute for %s err=%s\n", 
	     inet_2a(group, b1), strerror(errno));
      return -1;
    }
    return 1;
}

int get_sg_counts(struct mrt *mr)
{
  struct sioc_sg_req counts;
  
  counts.src.s_addr = mr->ms->src;
  counts.grp.s_addr = mr->mg->group;
  
  if (ioctl(igmp_sock, SIOCGETSGCNT, (char *) &counts) < 0) {
    printf("Error doing reading SG counts: %s\n", strerror (errno));
    return -1;
  }  

  mr->stats.old_packets = mr->stats.packets;
  mr->stats.old_bytes = mr->stats.bytes;
  mr->stats.old_time = mr->stats.time;

  mr->stats.packets = counts.pktcnt;
  mr->stats.bytes = counts.bytecnt;
  mr->stats.wrong_if = counts.wrong_if;
  mr->stats.time = time(NULL);
  return 1;
}

int add_to_oifs (char *name, vifi_t *oifs)
{
  listnode node;
  struct interface *ifp;

  for (node = listhead (iflist); node; nextnode (node))
    {
      ifp = getdata (node);
      if( strncmp(ifp->name, name, strlen(name)) == 0 ) {
		  VIFM_SET(ifp->ifindex, *oifs);
		  return 0;
      }      
    }
  printf("Warning intface found: %s\n", name);
   return 1;
}

void
select_DR()
{
  struct interface *ifp;
  struct pim_interface *pi;
  struct pim_neighbor *pn;
  listnode node, node_if;

  for (node_if = listhead (iflist); node_if; nextnode (node_if)) {
    ifp = getdata (node_if);
    
    pi = ifp->info;
    if(!pi) continue;

    /* If we don't have neigjbors we must be DR */
    pi->flags |= IF_DR;
	  
    for (node = listhead (pim_neighbor_list); node; nextnode (node)) {
      pn = getdata (node);
      if(pn->ifp == ifp ) {
	if( pi->address->u.prefix4.s_addr > pn->address.s_addr );
	else pi->flags &= ~IF_DR;
      }
    }
  }
}

void pim_flags_vty(struct vty *vty, u_int32_t flags)
{
	if(flags & MRT_LOCAL) vty_out(vty, "L");
	if(flags & MRT_CONNECTED) vty_out(vty, "C");
	if(flags & MRT_PRUNED) vty_out(vty, "P");
//	if(flags & MRT_WC) vty_out(vty, "WC");
	if(flags & MRT_RPT) vty_out(vty, "R");
	if(flags & MRT_SPT) vty_out(vty, "T");
	if(flags & MRT_SPARSE) vty_out(vty, "S");
	if(flags & MRT_DENSE) vty_out(vty, "D");
	if(flags & MRT_REGISTER) vty_out(vty, "F");
	if(flags & MRT_REGISTER_STOP) vty_out(vty, "[FS]");
	if(flags & MRT_BLACKHOLE) vty_out(vty, "B");
	if(flags & MRT_ASSERTED) vty_out(vty, "A");
	if(flags & MRT_PIM_IN_USE) vty_out(vty, "[PIM]");
	if(flags & MRT_IGMP_IN_USE) vty_out(vty, "[IGMP]");
}

void pim_flags_f(FILE *vty, u_int32_t s, u_int32_t g, u_int32_t flags)
{
  if(s)
    fprintf(vty, inet_2a(s, b1));
  else
    fprintf(vty, "*");
    fprintf(vty, ", %s  flags ", inet_2a(g, b1));
    
	if(flags & MRT_LOCAL) fprintf(vty, "L");
	if(flags & MRT_CONNECTED) fprintf(vty, "C");
	if(flags & MRT_PRUNED) fprintf(vty, "P");
//	if(flags & MRT_WC) fprintf(vty, "WC");
	if(flags & MRT_RPT) fprintf(vty, "R");
	if(flags & MRT_SPT) fprintf(vty, "T");
	if(flags & MRT_SPARSE) fprintf(vty, "S");
	if(flags & MRT_DENSE) fprintf(vty, "D");
	if(flags & MRT_REGISTER) fprintf(vty, "F");
	if(flags & MRT_REGISTER_STOP) fprintf(vty, "[FS]");
	if(flags & MRT_BLACKHOLE) fprintf(vty, "B");
	if(flags & MRT_ASSERTED) fprintf(vty, "A");
	if(flags & MRT_PIM_IN_USE) fprintf(vty, "[PIM]");
	if(flags & MRT_IGMP_IN_USE) fprintf(vty, "[IGMP]");
    fprintf(vty, "\n");
    fflush(vty);
}


void 
mr_list_oif(struct vty *vty, struct mrt *mr)
{
  struct oif *mi;
  listnode node;
  u_int32_t count = 0;

  for (node=listhead (mr->oif_list); node; nextnode (node)) {
    mi = getdata (node);
//    if( !(mi->flags & MRT_PIM_IN_USE)) continue;

    count++;
    vty_out(vty, "%s", VTY_NEWLINE);
  
    vty_out(vty, "  %s ", mi->ifp->name);
    uptime_vty (vty, &mi->uptime);
    vty_out(vty, "/");
    exptime_vty (vty, &mi->pim.exptime);
    vty_out(vty, " %s", inet_2a(mi->igmp.last_reporter, b1));
    
    vty_out(vty, " flags ");
    pim_flags_vty(vty, mi->flags);

  }
  if(!count) vty_out(vty, " Null");
  vty_out(vty, "%s", VTY_NEWLINE);
}

void
mroute_list_mrt(struct vty *vty, struct mgrp *mg, 
		struct mrt *mr, u_int32_t *flag )
{

   if(flag) {
      if(!(mr->flags & *flag )) return;
    }

    vty_out(vty, "%s", VTY_NEWLINE);

    vty_out(vty, " (%s, %s) ", inet_2a(mr->ms->src, b1), 
	    inet_2a(mr->mg->group, b2));
    uptime_vty (vty, &mr->uptime);
    vty_out(vty, "/");
    exptime_vty (vty, &mr->exptime);
    
    vty_out(vty, " flags ");
    pim_flags_vty(vty, mr->flags);
    
    vty_out(vty, "  %d/%d/%d%s", mr->stats.packets, mr->stats.bytes, 
	    mr->stats.wrong_if, VTY_NEWLINE);

    if(mr->flags & MRT_SPT)
    vty_out(vty, "  Incoming interface: %s,", &mr->ms->ifp->name? 
	    mr->ms->ifp->name:"Null");
    else
    vty_out(vty, "  Incoming interface: %s,", &mg->pim.rp->ifp->name? 
	    mg->pim.rp->ifp->name:"Null");

    vty_out(vty, " RPF nbr %s", inet_2a(mr->ms->rpf_neighbor, b1)); 
    
    vty_out( vty, "%s", VTY_NEWLINE );
    vty_out(vty, "  Outgoing interface list: ");
	
    if( list_isempty(mr->oif_list) ) {
      vty_out(vty, "Null%s", VTY_NEWLINE);
      return;
    }
    else {
      mr_list_oif(vty, mr);
    }
}

void 
mroute_list_mr_all(struct vty *vty, struct mgrp *mg, u_int32_t *flag)
{
  struct mrt *mr;
  listnode node;

  vty_out(vty, "%s", VTY_NEWLINE);

  for (node=listhead (mg->mrt_list); node; nextnode (node)) {
    mr = getdata (node);
    mroute_list_mrt(vty, mg, mr, flag );
  }
}

void 
mgrp_list_oif(struct vty *vty, struct mgrp *mg)
{
  struct oif *mi;
  listnode node;
  u_int32_t count = 0;

  for (node=listhead (mg->oif_list); node; nextnode (node)) {
    mi = getdata (node);
//    if( !(mi->flags & MRT_PIM_IN_USE)) continue;

    count++;

    vty_out(vty, "%s", VTY_NEWLINE);
    vty_out(vty, "  %s ", mi->ifp->name);
    uptime_vty (vty, &mi->uptime);
    vty_out(vty, "/");
    exptime_vty (vty, &mi->pim.exptime);
    vty_out(vty, " %s", inet_2a(mi->igmp.last_reporter, b1));
    vty_out(vty, " flags ");
    pim_flags_vty(vty, mi->flags);
  }
  if(!count) vty_out(vty, " Null");
  vty_out(vty, "%s", VTY_NEWLINE);
}

void
mroute_list_group(struct vty *vty, struct mgrp *mg, 
		  u_int32_t *mg_flag, u_int32_t *mr_flag)
{

     vty_out(vty, "%s", VTY_NEWLINE);
      vty_out(vty,"(*, %-s), ", inet_2a(mg->group,b1));
      uptime_vty (vty, &mg->pim.uptime);
      vty_out(vty, "/");
      exptime_vty (vty, &mg->pim.exptime);
      vty_out(vty, ", RP %-s  ",inet_2a(mg->pim.rp->src, b1)); 
      
      vty_out(vty, "flags ");
      pim_flags_vty(vty, mg->flags);
      vty_out(vty, "%s", VTY_NEWLINE);
      
      vty_out(vty, "  Incoming interface: %s,", &mg->pim.rp->ifp->name? 
	      mg->pim.rp->ifp->name:"Null");

      vty_out(vty, " RPF nbr %s%s", inet_2a(mg->pim.rp->rpf_neighbor, b1), 
	      VTY_NEWLINE );
      vty_out(vty, "  Outgoing interface list: ");
      if( list_isempty(mg->oif_list) ) {
	vty_out(vty, "Null%s", VTY_NEWLINE);
      }
      else {
	mgrp_list_oif(vty, mg);
      }
  
    if(!list_isempty(mg->mrt_list)) mroute_list_mr_all(vty, mg, mr_flag);
} 

void 
mroute_list(struct vty *vty, u_int32_t *mg_flag, u_int32_t *mr_flag)
{
  struct mgrp *mg;
  listnode node;

  for(node=listhead (mg_list); node; nextnode (node))
    {
      mg=getdata(node);
      
      if(mg_flag)
	{
	  if(!(mg->flags & *mg_flag )) continue;
	}
//      if( !(mg->flags & MRT_PIM_IN_USE)) continue;
      
      mroute_list_group(vty, mg, mg_flag, mr_flag);
    }
}

void 
mroute_list_active(struct vty *vty, u_int32_t *threshhold)
{
  listnode node, node1;
  time_t now, t;
  struct mgrp *mg;
  struct mrt *mr;
  struct sioc_sg_req counts;
  u_int32_t p_ave, p_kbits;
  u_int32_t default_threshhold = 4; /* Default 4 kbits */
  int i, print_group;

  if(!threshhold) threshhold = &default_threshhold;
  
  now = time(NULL);
  
  for(i=0;i <  MGRP_HASH_SIZE; i++) {
    if( mgrp_hash[i] == 0)  continue;
    for(node=listhead (mgrp_hash[i]); node; nextnode (node))
      {
	mg=getdata(node);
      	
	print_group = 1;
	for (node1=listhead (mg->mrt_list); node1; nextnode (node1)) {
	  mr = getdata (node1);
	  
	  counts.src.s_addr = mr->ms->src;
	  counts.grp.s_addr = mr->mg->group;
	  
	  if (ioctl(igmp_sock, SIOCGETSGCNT, (char *) &counts) < 0) {
	    printf("Error doing reading SG counts: %s\n", strerror (errno));
	    return;
	  }  
	  
	  t = now - mr->stats.time;
	  
	  /* This reacts faster than any EWMA.  :-) */

	  if(t < 4) {
	    t = now - mr->stats.old_time;
	    p_kbits =   (counts.bytecnt - mr->stats.old_bytes)/(t*128); 
	    p_ave  = (counts.pktcnt - mr->stats.old_packets)/t;
	  }
	  else {
	    p_kbits =   (counts.bytecnt - mr->stats.bytes)/(t*128); 
	    p_ave  = (counts.pktcnt - mr->stats.packets)/t;
	  }
	  
	  if(!t) continue;
	  if(p_kbits < *threshhold) continue;
	  
	  if(print_group) {
	    vty_out(vty, "** Active group %s%s",
		    inet_2a(mg->group, b1), VTY_NEWLINE);
	    print_group = 0;
	  }
	  
	  vty_out(vty, "  Source %s ", inet_2a(mr->ms->src, b1));
	  vty_out(vty, "%d kbit/s %d packets/s", 
		  p_kbits, p_ave);
	  
	  vty_out(vty, "%s", VTY_NEWLINE);
	}
      }
  }
}	


void
mroute_list_header(struct vty *vty)
{
 vty_out(vty, "IP Multicast Routing Table%s", VTY_NEWLINE);
 vty_out(vty, 
	 "Flags: A=Asserted, B=Blackholed, C=Connected, D=Dense, F=Register flag%s",
	 VTY_NEWLINE);
 vty_out(vty,
	 "  J=Join SPT, L=Local, P=Pruned, R=RP-bit set, S=Sparse, T=SPT-bit set%s",
	 VTY_NEWLINE);

 vty_out(vty,
	 "  [FS]= Register Stopped%s",
	 VTY_NEWLINE);


  vty_out(vty, "Timers: Uptime/Expires%s", VTY_NEWLINE);
}

DEFUN (show_ip_mroute,
       show_ip_mroute_cmd,
       "show ip mroute",
       "show ip mroute\n"
       "View multicast routes\n")
{
  mroute_list_header(vty);
  mroute_list(vty, NULL, NULL);
  return CMD_SUCCESS;
}

DEFUN (show_ip_mroute_actice,
       show_ip_mroute_active_cmd,
       "show ip mroute active",
       "show ip mroute active\n"
       "View active multicast routes\n")
{
  mroute_list_active(vty, NULL);
  return CMD_SUCCESS;
}

DEFUN (show_ip_mroute_flags,
       show_ip_mroute_flags_cmd,
       "show ip mroute flags WC-FLAGS SG-FLAGS",
       "show ip mroute filter\n"
       "View multicast routes with flags filter\n")
{

  u_int32_t mg_flag  = strtol (argv[0], NULL, 10);
  u_int32_t mr_flag  = strtol (argv[1], NULL, 10);
  mroute_list_header(vty);
  mroute_list(vty, &mg_flag, &mr_flag);
  return CMD_SUCCESS;
}

DEFUN (show_ip_mroute_filter_values,
       show_ip_mroute_filter_values_cmd,
       "show ip mroute filter-values",
       "show ip mroute filter-valus\n"
       "View multicast filter filter\n")
{

  mroute_list_header(vty);
  vty_out(vty, "L=%d, C=%d, P=%d, R=%d, T=%d, S=%d, D=%d, F=%d",
	  MRT_LOCAL, MRT_CONNECTED, MRT_PRUNED,  MRT_RPT,  MRT_SPT,
	  MRT_SPARSE, MRT_DENSE, MRT_REGISTER );

  vty_out(vty, "%s", VTY_NEWLINE);

  vty_out(vty, "[IGMP]=%d, [PIM]=%d, [REG_STP]=%d, B=%d, [ALL]=%d",
	  MRT_IGMP_IN_USE, MRT_PIM_IN_USE, MRT_REGISTER_STOP,  
	  MRT_BLACKHOLE, MRT_ALL);

  vty_out(vty, "%s", VTY_NEWLINE);

  return CMD_SUCCESS;
}


DEFUN (show_ip_mroute_group,
       show_ip_mroute_group_cmd,
       "show ip mroute A.B.C.D",
       "show ip mroute group\n"
       "View specific multicast route\n")
{
  struct in_addr group;
  struct mgrp *mg;
  int ret;

  ret = inet_aton (argv[0], &group);
  if (!ret) {
      vty_out (vty, "Malformed address %s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  if(!IN_MULTICAST(htonl(group.s_addr))) {
    vty_out (vty, "Not a multicast address %s", VTY_NEWLINE);
    return CMD_WARNING;
  }

  mroute_list_header(vty);

  if((mg=mg_lookup(group.s_addr)) == NULL ) return CMD_SUCCESS;

  mroute_list_group(vty, mg, NULL, NULL);
  return CMD_SUCCESS;
}

DEFUN (show_ip_mroute_mr,
       show_ip_mroute_mr_cmd,
       "show ip mroute A.B.C.D A.B.C.D",
       "show ip mroute (S,G) entry \n"
       "View specific multicast route\n")
{
  struct in_addr group, src;
  struct mgrp *mg;
  struct mrt *mr;
  int ret;

  ret = inet_aton (argv[0], &group);
  if (!ret) {
      vty_out (vty, "Malformed multicast address %s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  if(!IN_MULTICAST(htonl(group.s_addr))) {
    vty_out (vty, "Not a multicast address %s", VTY_NEWLINE);
    return CMD_WARNING;
  }


  ret = inet_aton (argv[1], &src);
  if (!ret) {
      vty_out (vty, "Malformed unicast address %s", VTY_NEWLINE);
      return CMD_WARNING;
    }

  if(IN_BADCLASS(htonl(src.s_addr))) {
    vty_out (vty, "Not a multicast address %s", VTY_NEWLINE);
    return CMD_WARNING;
  }

  mroute_list_header(vty);

  if((mg=mg_lookup(group.s_addr)) == NULL ) return CMD_SUCCESS;
  if((mr=mr_lookup(src.s_addr, group.s_addr)) == NULL ) return CMD_SUCCESS;

  vty_out(vty, "%s", VTY_NEWLINE);
  mroute_list_mrt(vty, mg, mr, NULL);
  return CMD_SUCCESS;
}


int ms_list(struct vty *vty)
{
  struct msrc *s;
  struct in_addr a;
  listnode node;
  int i;
  
  for(i=0;i <  MSRC_HASH_SIZE; i++) {
    if( msrc_hash[i] == 0)  continue;
    for (node = listhead(msrc_hash[i]); node; nextnode(node)) {
      s = getdata( node );
      if(s) {
        a.s_addr = s->src;
	vty_out(vty,"%-16s", inet_ntoa(a));
	vty_out(vty, " ");
	
	uptime_vty (vty, &s->uptime);
	vty_out(vty, "  ");
	vty_out(vty, "     %-16s", inet_2a(s->rpf_neighbor, b1));
	
	vty_out(vty, "  %-8s  %7u", s->ifp->name, s->use_count);
	vty_out(vty, "  [%lu/%lu]", s->preference, s->metric);
	vty_out(vty, "%s", VTY_NEWLINE);
      }
    }
  }
  return 0;
}

DEFUN (show_ip_multicast_sources,
       show_ip_multicast_sources_cmd,
       "show ip multicast sources",
       "show ip multicast sources\n"
       "View multicast multicast sources\n")
{

 vty_out(vty, "IP Multicast Source Table%s", VTY_NEWLINE);

 ms_list(vty);
  return 1;
}

void
route_cmd_init()
{
  install_element (VIEW_NODE, &show_ip_mroute_cmd); 
  install_element (VIEW_NODE, &show_ip_mroute_group_cmd); 
  install_element (VIEW_NODE, &show_ip_mroute_mr_cmd); 
  install_element (VIEW_NODE, &show_ip_mroute_active_cmd); 
  install_element (VIEW_NODE, &show_ip_mroute_flags_cmd);
  install_element (VIEW_NODE, &show_ip_mroute_filter_values_cmd);
  install_element (VIEW_NODE, &show_ip_multicast_sources_cmd); 
  install_element (ENABLE_NODE, &show_ip_mroute_cmd);
  install_element (ENABLE_NODE, &show_ip_mroute_mr_cmd); 
  install_element (ENABLE_NODE, &show_ip_mroute_active_cmd); 
  install_element (ENABLE_NODE, &show_ip_mroute_flags_cmd);
  install_element (ENABLE_NODE, &show_ip_mroute_filter_values_cmd);
  install_element (ENABLE_NODE, &show_ip_mroute_group_cmd); 
  install_element (ENABLE_NODE, &show_ip_multicast_sources_cmd); 
}

